import weakref
from asyncio import Event
from queue import Queue
import functools

class SharableAsyncIterable:

    def __init__(self, async_iterable):
        self._async_iterable = async_iterable
        self._async_iterator = None
        self._weak_ref_child_iterators = []
        self._noone_broadcasting = Event()
        self._noone_broadcasting.set()

    async def _broadcast(self):
        value = await self._async_iterator.__anext__()
        for ref in list(self._weak_ref_child_iterators):
            child = ref()
            if child is None:
                self._weak_ref_child_iterators.remove(ref)
            else:
                child._queue.put(value)

    def __aclose__(self):
        print("what do you know")


    def __aiter__(self):

        class _SharableAsyncIterator:

            def __init__(self, parent):
                self._parent = parent
                self._queue = Queue()

            def __aclose__(self):
                print("another what do you know")

            async def __anext__(self):
                if not self._queue.empty():
                    return self._queue.get()
                else:
                    if self._parent._noone_broadcasting.is_set():
                        self._parent._noone_broadcasting.clear()
                        await self._parent._broadcast()
                        self._parent._noone_broadcasting.set()
                    else:
                        await self._parent._noone_broadcasting.wait()
                    return self._queue.get()

        shar_iter = _SharableAsyncIterator(self)
        weak_ref_ret = weakref.ref(shar_iter)
        if not self._weak_ref_child_iterators:
            self._async_iterator = self._async_iterable.__aiter__()
        self._weak_ref_child_iterators.append(weak_ref_ret)

        return shar_iter

class ReplayAsyncIterable:

    def __init__(self, async_iterator):
        self._async_iterator = async_iterator
        self._values = []
        self._noone_broadcasting = Event()
        self._noone_broadcasting.set()

    async def _broadcast(self):
        self._values.append(await self._async_iterator.__anext__())


    def __aiter__(self):

        class _ReplayAsyncIterator:

            def __init__(self, parent):
                self._parent = parent
                self._count = 0

            async def __anext__(self):
                if self._count < len(self._parent._values):
                    value = self._parent._values[self._count]
                    self._count += 1
                    return value
                else:
                    if self._parent._noone_broadcasting.is_set():
                        self._parent._noone_broadcasting.clear()
                        await self._parent._broadcast()
                        self._parent._noone_broadcasting.set()
                    else:
                        await self._parent._noone_broadcasting.wait()
                    value = self._parent._values[self._count]
                    self._count += 1
                    return value

        shar_iter = _ReplayAsyncIterator(self)

        return shar_iter



import functools
# from typing import AsyncIterable, AsyncIterator

from .tools import SharableAsyncIterable, ReplayAsyncIterable


def share(obj):
    if callable(obj):
        return _async_share_dec(obj)
    elif hasattr(obj, "__aiter__"):
        return _async_share(obj)
    else:
        raise Exception("share can only be applied to an AsyncIterable object.")


def repeat(obj):
    if callable(obj):
        return _async_replay_dec(obj)
    elif hasattr(obj, "__anext__"):
        return _async_replay(obj)
    else:
        raise Exception("repeat can only be applied to an AsyncIterator object.")


def _async_share(async_iterable):
    return SharableAsyncIterable(async_iterable)


def _async_replay(async_iterator):
    return ReplayAsyncIterable(async_iterator)


def _async_share_dec(async_iterable_fn):
    def wrapper(*args, **vargs):
        obj = async_iterable_fn(*args, **vargs)
        if not hasattr(obj, "__aiter__"):
            raise Exception("share can only be applied to an AsyncIterable object.")
        return SharableAsyncIterable(obj)
    return functools.cache(wrapper)


def _async_replay_dec(async_iterator_fn):
    def wrapper(*args, **vargs):
        obj = async_iterator_fn(*args, **vargs)
        if not hasattr(obj, "__anext__"):
            raise Exception("repeat can only be applied to an AsyncIterator object.")
        return ReplayAsyncIterable(obj)
    return functools.cache(wrapper)



from distutils.core import setup

setup(
    name="rxiter",
    license="MIT",
    version="0.0.2",
    packages=["rxiter"],
    install_requires=["functools"],
    description="Observable operations for async generators",
    classifiers=[
        "Development Status :: 6 - Mature",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python"
    ]
)
